/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>

#include "pool.h"
#include "dist.h"



typedef unsigned int packed;

typedef struct ocr_data_entry {
	char color[10];
	packed *point;
} ocr_data_entry;

int num_points;

ocr_data_entry *ocr_data;

size_t ocr_data_alloc;
size_t ocr_poolnum;


int	image_size,
	len_block = sizeof(packed)*8,
	packed_len;


void binary_print(unsigned int i) {

	int j;
	for (j=len_block-1;j>=0;j--) printf("%d",(i>>j)&01);
	printf(" ");
}


void this_show_point(size_t i) {

	int j;
	printf("%d %s ",i, ocr_data[i].color);
	for (j=0;j<packed_len;j++) binary_print(ocr_data[i].point[j]);
	printf("\n");
}

void pack_im(char *inl, unsigned int *tv) {
	unsigned int i;
	int j;
	for (i=0; (int)i<image_size; i++) {
		j = i/len_block;
		tv[j]<<=1;
		tv[j] += inl[i] - '0';
	}
}



int same_color(int i,int j) {
	if ((i<0) || (j<0)) return 0;
	return strcmp(ocr_data[i].color,ocr_data[j].color)==0;
}



#define MAX_BCOUNT 65536
#define LEN_BCOUNT 16

unsigned int bit_count[MAX_BCOUNT];

void init_bitcount(void) {

	int i,j;
	unsigned int v;
	for (i=0; i<MAX_BCOUNT; i++)
		for (j=0,v=i;j<LEN_BCOUNT;j++,v>>=1)
			bit_count[i] += v & 01;
}



float this_dist(size_t i, size_t j) {

	int k,d;
	unsigned int	*a = ocr_data[i].point,
	    		*b = ocr_data[j].point,
			c;

	if (!a || !b) abort();

	for (d=k=0; k<packed_len; k++) {
		c = a[k]^b[k];
		d += bit_count[c>>LEN_BCOUNT] + bit_count[c&0x0000ffff];
							 /* mask is MAX_BCOUNT-1 */
						/* only works because LEN_BCOUNT==len_block/2 */
	}
	return (float)d;
}

float this_dist2(size_t i, size_t j) {return this_dist(i, j);}


size_t get_data(FILE *IN) {

	int nr=0, nc=0, nrt, nct;
	char inl[4000];

	num_points=0;
	init_bitcount();
	ocr_data_alloc = 3000;
	ocr_data = (ocr_data_entry*)calloc(ocr_data_alloc, sizeof(ocr_data_entry));

	while (fscanf (IN, "%s%d%d%s", ocr_data[num_points].color, &nrt, &nct, inl) != EOF) {
		if (nr==0 && nc==0) {
			nr = nrt;
			nc = nct;
			image_size = nr*nc;
			packed_len = (image_size-1)/len_block + 1;
			ocr_poolnum = pool_init(packed_len*sizeof(packed));
		} else if (nr!=nrt || nc != nct || (int)strlen(inl)!=image_size)
			abort();
		if (num_points>(int)ocr_data_alloc-2) {
			ocr_data_alloc *= 2;
			ocr_data=(ocr_data_entry*)realloc(ocr_data, ocr_data_alloc*sizeof(ocr_data_entry));
		}
		ocr_data[num_points].point = (packed*)pool_new(ocr_poolnum);
		pack_im(inl, ocr_data[num_points++].point);
	}
	return num_points;
}

void free_data(void){}

size_t this_color(int x) {return 0;}

