/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdlib.h>
#include <stdarg.h>
#include <math.h>

#include "dist.h"
#include "my_rand.h"
#include "rand_gen.h"

/* 
	adapted from rand.cc of Arya and Mount,
	available via http://www.cs.umd.edu/~mount/ANN/
*/


double rand_laplace(double dum) 
{
	double r, x;
	do r = my_rand();
	while (r==0.0);
	x = -log(r) / 1.41421356237;
	return (my_rand() < 0.5) ? x : -x;
}


double rand_uniform(double dum) {return 2*my_rand()-1;}


double norm22(coord *p) { return p[0]*p[0] + p[1]*p[1];}
double rand_gauss(double std_dev) {
	coord v[2];
	double s;
	while ((s=norm22(rand_uniform_pt(v, 2, 0)))>1) {}
	return std_dev*v[0]*sqrt(-2*log(s)/s);
}



coord *ind_rand_pt(coord *p, int dim, double rand_param, double (*f)(double))
{
	int d;
	for (d=0; d<dim; d++) p[d] = f(rand_param);
	return p;
}


coord *rand_gauss_pt(coord *p, int dim, double std_dev)
	{return ind_rand_pt(p, dim, std_dev, rand_gauss);}
coord *rand_laplace_pt(coord *p, int dim, double dum)
	{return ind_rand_pt(p, dim, 0, rand_laplace);}
coord *rand_uniform_pt(coord *p, int dim, double dum)
	{return ind_rand_pt(p, dim, 0, rand_uniform);}

coord *rand_co_gauss_pt(coord *p, int dim, double correlation) {
	int d;
	double std_dev_w = sqrt(1.0 - correlation * correlation);
	p[0] = rand_gauss(1);
	for (d=1; d<dim; d++)
		p[d] = correlation*p[d-1] + std_dev_w*rand_gauss(1);
	return p;
}

coord *rand_co_laplace_pt(coord *p, int dim, double correlation) {
	int d;
	double corr_sq = correlation*correlation;
	p[0] = rand_laplace(0);
	for (d=1; d<dim; d++)
		p[d] = correlation*p[d-1]
		   +  ((my_rand() < corr_sq) ? 0 : rand_laplace(0));
	return p;
}

