/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <math.h>

#include "my_rand.h"
#include "gen_dist.h"
#include "pool.h"
#include "sb.h"
#include "Msbi.h"
#include "MMM.h"
#include "do_test.h"
#include "dist.h"
#include "rand_gen.h"

type_data this_one;

FILE *popen(char*, char*);

coord	**P;

#define N_ELEMS(x) ((int)(sizeof(x)/sizeof((x)[0])))

int dims[] = {2, 4, 8, 16, 24, 32};
int dim_pool[30];
int num_sites[] = {256, 512, 1024, 2048, 4096, 8192, 16384, 32768};
int Ks[] = {100};
float gammas[] = {1.2};
int Qsvns[] = {2, 4};

static void generate_data_r(
  coord **P, int i_dims, int num_tot, rand_pt_f rr, double rr_param) {
	int i,j, dim = dims[i_dims];

	for (j=0; j<num_tot; j++) {
		P[j] = (coord*)pool_new(dim_pool[i_dims]);
		for (i=0;i<dim;i++) rr(P[j], dim, rr_param);
	}
	this_one.P = P;
	this_one.dim = dim;
	this_one.num_P = num_tot;
}

static void gen_uniform_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_uniform_pt, 0);}

static void gen_laplace_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_laplace_pt, 0);}

static void gen_gauss_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_gauss_pt, 1);}

static void gen_co_laplace_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_co_laplace_pt, 0.8);}

static void gen_co_gauss_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_co_gauss_pt, 0.8);}


int num_centers = 10;
static void gen_clus_gauss_data(int i_dims, int num_tot) {
	int i,j, dim = dims[i_dims];
	coord **centers = (coord**)calloc(num_centers+1, sizeof(coord*));
;

	generate_data_r(centers, i_dims, num_centers, rand_uniform_pt, 0);
	generate_data_r(P, i_dims, num_tot, rand_gauss_pt, 1.0/2/num_centers);
		/* this_one is now correctly set */

	for (j=0; j<num_tot; j++) {
		int jp = my_rand()*num_centers;
		for (i=0;i<dim;i++) P[j][i] += centers[jp][i]; 
	}

	if (0) for (j=0; j<num_centers; j++) pool_free(dim_pool[i_dims], &centers[j]);
}

typedef void gen_data_f(int, int);

struct {char *nam; gen_data_f* func;}
	gen_data_fs[] = { {"uniform_data", &gen_uniform_data}
/*
			,{"laplace_data", &gen_laplace_data}
			,{"gauss_data", &gen_gauss_data}
			,{"co_gauss_data", &gen_co_gauss_data}
			,{"co_laplace_data", &gen_co_laplace_data}
			,{"clus_gauss_data", &gen_clus_gauss_data}
*/
		};


main(int argc, char** argv) {

	size_t i_dims, i_num_sites, i_Ks, i_gammas, i_Qsvns, i_gen_data_fs;
	size_t trial_count=0, cc=0;
	char packing = 1;
	size_t nt, ntd=100, L=5;
	float end_frac=0.1, num_reps = 5;
	sbp sbt;
	size_t i, maxn;
	size_t num_P;

	fscanf(popen("wc test_M_run", "r"), "%d", &trial_count);
	printf("trialcount/num_reps=%d\n", trial_count/num_reps);

	switch (argc) {
		case 6: packing = atoi(argv[5]);
		case 5: num_reps = atoi(argv[4]);
		case 4: ntd = atoi(argv[3]);
		case 3: packing = atoi(argv[2]);
		case 2:	end_frac= atof(argv[1]);	/* intentional fallthru */
	}

	init_my_rand();

	maxn = ntd+num_sites[(N_ELEMS(num_sites)-1)]*Qsvns[(N_ELEMS(Qsvns)-1)]+ntd;
	shufmat = (size_t*)malloc(maxn*sizeof(size_t));
	for (i=0; i<maxn; i++) shufmat[i] = i;

	P = (coord**)calloc(maxn+1, sizeof(coord*));

	ITERATE(dims) {
		dim_pool[i_dims] = pool_init(dims[i_dims]*sizeof(coord));
		ITERATE(num_sites) ITERATE(Ks) ITERATE(gammas) ITERATE(Qsvns) ITERATE(gen_data_fs) {
			int L = Qsvns[i_Qsvns];
			if (cc++ < trial_count) continue;
			num_P = num_sites[i_num_sites]*L;
			nt = ntd;
			printf("%d	%d	%d	%s",
				num_P/L, dims[i_dims], Qsvns[i_Qsvns],
					gen_data_fs[i_gen_data_fs].nam);
			gen_data_fs[i_gen_data_fs].func(i_dims, num_P+nt);
if (0 && gen_data_fs[i_gen_data_fs].func==gen_co_laplace_data) {
	FILE *debclus = fopen("debclus", "w");
	int i,j;
		for (j=0; j<num_P+nt; j++) {
			for (i=0;i<dims[i_dims];i++) fprintf(debclus, " %G", P[j][i]);
			fprintf(debclus, "\n");
		}
	fclose(debclus);
}

			use_data(&this_one);
			memset(dcounts, 0, sizeof(dcounts));
			sbt = build_Mp(my_dist, num_P/L, num_P, Ks[i_Ks],
					gammas[i_gammas], end_frac, 0, packing);
			print_prep(sbt);

			test_sbt(sbt, num_P/L, num_P, nt, 0, 0 , search_M);
			free_sb(sbt,0);
			fflush(stdout);
		}
		pool_purge(dim_pool[i_dims]);
	}
	pool_purge_all();
	return 0;
}
