/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


/**************************************************************************
*	search structure M
*	
*	build using build_M, build_Mp, or build_M_simple;
*	search using search_M
*
*	points 0 up to num_sites-1 should be sites, for which a nearest
*	neighbor structure is built; points num_sites up to num_tot-1
*	should be the "training query points".
*
*	data structure size and answer accuracy increase in K and gamma,
*	query time decreases with end_frac, storage and prep. decrease with
*	end_frac
*
**************************************************************************/

typedef size_t (*color_func)(size_t i);

sbp build_M(
	sb_distance	
	,size_t num_sites	/* number of sites */
	,size_t num_tot		/* num_sites + number of "training" query points*/
	);


				/* version with more parameters */
sbp build_Mp(
	sb_distance	
	,size_t num_sites	/* number of sites */
	,size_t num_tot		/* num_sites + number of "training" query points */
	,size_t K		/* default 1000; K>= log(num_sites) seems ok */
	,float gamma		/* default 1.2 */
	,float end_frac		/* default 0.1 */
	,color_func col 	/* if nonzero, data structure returns point with
					same color as NN */
	,char packing		/* if 0, look at sites in random order;
					otherwise, "pack" by picking
					site farthest from sites
					currently inserted
				   default 1
				*/
	);



		/* same calling sequence as build_Mp, quadratic but 
		   does not need triangle inequality for correctness */

sbp build_Mp_simple(sb_distance dist, size_t num_sites, size_t num_tot,
	size_t K, float gamma, float end_frac, color_func col, char packing);



/**************************************************************************
* given sbp built by build_M, and point number q,
*	returns likely nearest site to q
***************************************************************************/

typedef size_t search_M_f(sbp, size_t);

search_M_f search_M, search_M_simple;

