/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


/********************************************
 * M structure
 ********************************************/

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "pool.h"
#include "heap.h"
#include "sb.h"
#include "MMM.h"
#include "Msbi.h"
#include "do_test.h"

static char	show_progress = 0,
	show_sb_flag = 0;

static int alist_str_poolnum = -1;

static SD q_nn;

static heap sh;

typedef struct site_data_M_str {
	SD nn;
	size_t last_touching;
	char multicolor;
	size_t offset_thresh;
	alist affL;
} site_data_M_str;



sb *build_Mp_simple(sb_distance dist, size_t num_sites, size_t num_tot,
	size_t K, float gamma, float end_frac, color_func col, char packing)
{
    size_t i, j, Kimax, Ki;
    float dji;
    SD *nnj, fnn;
    alist iaffL, v;
    sb *sbt;
    site_data_M_str *sd_M;
    int ii;

	activity = prep;

    if (alist_str_poolnum==-1)
        alist_str_poolnum = pool_init(sizeof(alist_str));
    sbt = init_sb(dist, num_sites, end_frac, 0, 0);
    sbt->col = 0;

    assert(dist!=0 && K>0 && gamma >=1);

    if ((Kimax = (K+1)*num_sites) > num_tot)
        Kimax = num_tot;

    make_storage(sbt->ins_num, size_t, sbt->n);
    for (i=0; i<sbt->n; i++) sbt->ins_num[i] = i;

    sd_M = (site_data_M_str*)calloc(num_tot, sizeof(site_data_M_str));
    for (j=0; j<num_sites; j++)
        sd_M[j].last_touching = num_tot+1;
    for (j=0; j<Kimax; j++) {
        sd_M[j].nn.point = 0;
        sd_M[j].nn.key = (*dist)(0, j);
    }

	for (Ki = K, ii=1; ii<(int)num_sites; ii++, Ki+=K) {
		if (packing) for (fnn.key=0, j=0; j<num_sites; j++) {
			if (fnn.key < sd_M[j].nn.key) {
				fnn.key = sd_M[j].nn.key; fnn.point = j;
			}
		} else fnn.point = ii;
		sbt->ins_num[i = fnn.point] = ii;
		if (ii==(int)num_sites-1)
			printf("gap is %g\n", fnn.key);
		
		if (show_progress && i%100==0) {printf(".");fflush(stdout);}
		for (j=0; j < Kimax; j++) {
			nnj = &sd_M[j].nn;
			dji = (*dist)(i,j);
			if (dji >= nnj->key*gamma) continue;
			if (j < Ki && sd_M[nnj->point].last_touching!=i && (int)i!=nnj->point) {
				/* i touchs nnj->point */
				sd_M[nnj->point].last_touching = i;
				v = (alist)pool_new(alist_str_poolnum);
				sbt->offset[nnj->point+1]++;
				v->site = i;
				v->next = sd_M[nnj->point].affL;
				sd_M[nnj->point].affL = v;
			}
			if (i>=sbt->end_thresh || dji >= nnj->key) continue;
			nnj->key = dji;
			nnj->point = i;
		}
	}


    for (i=1; i<=sbt->n;i++) {
        sbt->offset[i] += sbt->offset[i-1];
    }

    make_storage(sbt->DL, size_t, sbt->offset[sbt->n]);

    for (i=0; i<sbt->end_thresh; i++) {
        iaffL = sd_M[i].affL;
        for (j=0; j<sbt->offset[i+1]-sbt->offset[i]; j++) {
            assert(iaffL);
            sbt->DL[sbt->offset[i+1]-1-j] = iaffL->site;
           iaffL = iaffL->next;
        }
        assert(!iaffL);
    }

    pool_purge(alist_str_poolnum);
    free(sd_M);
    if (show_sb_flag) show_sb(sbt);
    return sbt;
}

sbp build_M(sb_distance dist, size_t num_sites, size_t num_tot) {
	return build_Mp(dist, num_sites, num_tot, 1000, (float)1.2, (float)0.1, 0, 1);
}

/***********************************************************************
 *    for storing DLs as found, used in build_M
 ***********************************************************************/


typedef struct edge {int a, b; float d;} edge;

edge *edges=0;
size_t    num_edges, alloc_edges=0;

sb *sbt_cmp;

int cmp_edges(const void *x, const void *y)
{ 
    edge ex = *(edge*)x,
         ey = *(edge*)y;
    size_t exa = ex.a,
           exb = sbt_cmp->ins_num[ex.b],
           eya = ey.a,
           eyb = sbt_cmp->ins_num[ey.b];
        
    if (exa < eya) return -1;
    if (exa > eya) return 1;
    if (exb < eyb) return -1;
    if (exb > eyb) return 1;
    return 0;
}


static void compress_edges(sb* sbt) {
    size_t i, j;

    if (num_edges==0) return;
    sbt_cmp = sbt;
    qsort(edges, num_edges, sizeof(*edges), cmp_edges);

    for (i=0, j=1; j<num_edges; j++)
	if (edges[i].a!=edges[j].a || edges[i].b!=edges[j].b)
		edges[++i] = edges[j];

    num_edges = i+1;

}

static void insert_edge(sb *sbt, size_t nnew, size_t touched) {

    if (num_edges>=alloc_edges) {
	alloc_edges = 1 + 2*alloc_edges;
        edges = (edge*)realloc(edges, sizeof(edge)*alloc_edges);
		assert(edges);
        compress_edges(sbt);
    }
    edges[num_edges].a = touched;
    edges[num_edges].b = nnew;
    edges[num_edges].d = sbt->dist(touched, nnew);
    num_edges++;
}



sbp build_Mp(
	sb_distance dist, size_t num_sites, size_t num_tot, size_t K,
	float gamma, float end_frac, color_func col, char packing)
{
	
	
    size_t  m, ma, q;
    int i, j, k;
    float d;
    sb *sbt, *sb_sbt = build_sbp(dist, num_sites, 1, packing, 0);
	
    assert(dist!=0 && K>0 && gamma >=1);

	activity = prep;
	
    num_edges = 0;
	
    sbt = init_sb(dist, num_sites, end_frac, 0, 0);
    sbt->col = col;
    make_storage(sbt->ins_num, size_t, sbt->n);
    memcpy(sbt->ins_num, sb_sbt->ins_num, sbt->n*sizeof(size_t));
    init_heap(&sh);
	
    sb_sbt->ns = 0;
    for (q=0; q<num_tot; q++) {
        q_nn.key = FLT_MAX;
        j = ma = 0;
        d = 0;
        while (1) {
            if (  !sb_sbt->seen_for[j]
				&& sb_sbt->nn[ma].max.old + gamma*q_nn.key > d     
				) {
                sb_sbt->dist2q[j] = d = sb_sbt->dist(q,j);
                if (d<gamma*q_nn.key
					&& q < sb_sbt->n + K*sbt->ins_num[j] && j!=0
					&& sbt->ins_num[j] > sbt->ins_num[q_nn.point]
					) {
                    insert_edge(sbt, j, q_nn.point);
                }
                if (d < q_nn.key
                    && sbt->ins_num[j] < sbt->end_thresh) {
                    q_nn.key = d; q_nn.point = j;
				}
                ma = sb_sbt->cur_touching[j] = sb_sbt->offset[j];
                if (ma<sb_sbt->offset[j+1]) {
                    int Z=-(int)sb_sbt->ins_num[sb_sbt->DL[ma]];
                    insert_heap(&sh, j, (float)Z);
                    sb_sbt->seen_for[j] = 1;
                    sb_sbt->seens[sb_sbt->ns++] = j;
                }
            }
            if (sh.num_entries==0) break;
            do {
                stat_inc_search_loop;
                m = sh.h[0].point;
                d = sb_sbt->dist2q[m];
                ma = (sb_sbt->cur_touching[m])++;
                delete_heap(&sh, 0);
                if (ma>=sb_sbt->offset[m+1]
					|| sb_sbt->dDL[ma] + q_nn.key < d 
					) continue;
                j = sb_sbt->DL[ma];
                if (ma+1<sb_sbt->offset[m+1]) {
                    int Z=-(int)sb_sbt->ins_num[sb_sbt->DL[ma+1]];
                    insert_heap(&sh, m, (float)Z);
                }
                stat_update_max_heap_entries(sh.num_entries);
                break;
            } while (sh.num_entries>0);
        }
        reset_seens(sb_sbt);
        sh.num_entries = 0;
    }
	
    compress_edges(sbt);
	
    if (sbt->col) {
        printf("adjusting for color\n");
        for (i=j=num_edges-1; j>=0;) {
            while (j>= 0 && edges[j].a==edges[i].a
                && sbt->col(edges[j].b)==sbt->col(edges[i].a)
				)
                edges[j--].b = -1;
            while (j>=0 && edges[j].a==edges[i].a) j--;
            i = j;
        }
        for (i=j=0; j<(int)num_edges; j++) if (edges[j].b!=-1) edges[i++]=edges[j];
        num_edges = i;
    }
	
    make_storage(sbt->DL, size_t, num_edges);
    make_storage(sbt->dDL, float, num_edges);
	
    sbt->offset[0] = 0;
    for (i=j=0; j<(int)num_edges; j++) {
        sbt->DL[j] = edges[j].b;
		sbt->dDL[j] = edges[j].d;
        if (edges[j].a!=edges[i].a)
            for (k=edges[i].a+1; k<=edges[j].a; k++) {
                sbt->offset[k] = j;
                i = j;
            }
    }
	
    for (k=edges[i].a+1; k<=(int)sbt->n; k++)
		sbt->offset[k] = num_edges;
	
    if (show_sb_flag) show_sb(sbt);
	
    free(sh.h);
    free_sb_i(sb_sbt,1);
    return sbt;
	
}

/***********************************************************************
 * search M structure
 ***********************************************************************/

size_t search_M(sbp sbt, size_t q) {

    size_t t=0, s, i, h, r=0;
    float tdist, idist, rdist;

	activity = query;

    rdist = tdist = sbt->dist(q,0);
    do {
        h = sbt->offset[t+1];
        for (s=sbt->offset[t]; s<h; s++) {
            i = sbt->DL[s];
            if (sbt->dDL[s] > tdist + rdist) continue; 
            idist = sbt->dist(q,i);
            if (idist >= rdist) continue;
            r = i;
            rdist = idist;
            if (sbt->ins_num[r] >= sbt->end_thresh) continue;
            t = i;
            tdist = idist;
            break;
        }
    } while (s!=h);
    return r;
}


/***********************************************************************
 * search M structure, non-metric; no sbt->dDL
 ***********************************************************************/

size_t search_M_simple(sbp sbt, size_t q) {

    size_t t=0, s, i, h, r=0;
    float  idist, rdist;

	activity = query;

    rdist = sbt->dist(q,0);
    do {
        h = sbt->offset[t+1];
        for (s=sbt->offset[t]; s<h; s++) {
            i = sbt->DL[s];
            idist = sbt->dist(q,i);
            if (idist >= rdist) continue;
            r = i;
            rdist = idist;
            if (sbt->ins_num[r] >= sbt->end_thresh) continue;
            t = i;
            break;
        }
    } while (s!=h);
    return r;
}
