/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include "stdio.h"
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <string.h>


#include "my_rand.h"
#include "gen_dist.h"
#include "sb.h"
#include "Msbi.h"
#include "MMM.h"
#include "do_test.h"

double alphas[] = {0.1, 0.25, 0.5, 0.75, 0.9, 1};
/* float alphas[] = {1};*/

struct {char *nam; search_sb_f* func;}
	search_sbs[] = { {"d_order", &search_sb_d_order}
			,{"i_order", &search_sb_i_order}
			,{"stack", &search_sb_stack}
			,{"semi_h", &search_sb_semi_heap}
			,{"no_h", &search_sb_no_heap}
		};

main(int argc, char* argv[]) {

	sbp sbt;
	char packing = 1;
	size_t nt, i_alphas, i_search_sbs; 
	float end_frac=(float)0.1;
	FILE *IN = (argc>=2) ? fopen(argv[1], "r") : stdin;
	int num_P, L=5;

	if (IN==0)
		exit(2);

	num_P = get_data(IN);

	fprintf(stderr, "got data\n"); fflush(stderr);

	if (num_P==0){
		fprintf(stderr, "empty data file?");
		exit(0);
	}


	switch (argc) {
		case 4: packing = atoi(argv[3]);
		case 3:	end_frac= (float)atof(argv[2]);	/* intentional fallthru */
	}


	init_my_rand();

	if (num_P/L < 500) nt = num_P/L; else nt = 500;

	make_shuffle(num_P, num_P-nt, num_P);

	dcounts[prep] = 0;
	activity = prep;
	sbt = build_sbp(my_dist, num_P-nt, end_frac, packing, my_dist2);

	printf("dataset=%s\n", (IN==stdin) ? "stdin" : argv[1]);
	print_prep(sbt);

	printf("num_sites=%d\n", num_P-nt);

	ITERATE(search_sbs) {
		printf("search_method=%s\n", search_sbs[i_search_sbs].nam);
		ITERATE(alphas) {
			printf("alpha=%0.1f\n", alphas[i_alphas]);
			test_sbt(sbt, num_P-nt, num_P-nt, nt, (float)alphas[i_alphas],
				search_sbs[i_search_sbs].func, 0);
		}
	}

	free_data();

	free_sb_i(sbt,1);

	return 0;

}
