/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include "stdlib.h"
#include "assert.h"
#include "string.h"

#define MAXBLOCKS 10000
#define MAXPOOLS 20
#define Nobj 1000

typedef struct block_el {
	int ref_count;
	struct block_el *next;
	char stuff[1];
} block_el;

static block_el *pool_blocks[MAXPOOLS][MAXBLOCKS];

typedef struct pool_el {
	size_t size;
	int num_blocks;
	block_el *free_list;
} pool_el;

static pool_el pools[MAXPOOLS];
static int	num_pools = 0;

int pool_init(size_t m) {
	pool_el* pn;
	assert(num_pools < MAXPOOLS);
	pn = &pools[num_pools];
	pn->num_blocks = 0;
	pn->size = m;
	pn->free_list = 0;
	return num_pools++;
}

static block_el *pool_new_block(int poolnum) {
	pool_el* pn = &pools[poolnum];
	block_el *xbt;
	int i;

	assert(poolnum>=0 && poolnum<num_pools);
	assert(pn->num_blocks < MAXBLOCKS);

	xbt = pool_blocks[poolnum][pn->num_blocks] =
		(block_el*)calloc(Nobj, pn->size);

	assert(xbt);

	for (i=0; i<Nobj-1; i++) {
		xbt->next = (block_el*)((char*)xbt + pn->size);
		xbt = xbt->next;
	}
	return pool_blocks[poolnum][pn->num_blocks++];
}

void *pool_new(int poolnum) {
	pool_el* pn = &pools[poolnum];
	block_el *q = pn->free_list ? pn->free_list : pool_new_block(poolnum);
	assert(q);
	pn->free_list = q->next;
	memset(q, 0, pn->size);
	return (void*)q;
}

void pool_free(int poolnum, void **x) {
	int i;
	pool_el* pn = &pools[poolnum];
	block_el *q = (block_el*) *x;
	for (i=0; i<pn->num_blocks; i++)
		if (q - pool_blocks[poolnum][i] < Nobj
			&& q - pool_blocks[poolnum][i] > 0) break;
	assert(i!=pn->num_blocks);
	q->next = pn->free_list;
	pn->free_list = q;
	*x = 0;
}


void pool_purge(int poolnum) {
	int i;
	pool_el* pn = &pools[poolnum];

	for (i=0; i<pn->num_blocks; i++)
		free(pool_blocks[poolnum][i]);
	pn->num_blocks = 0;
	pn->free_list = 0;
}


void pool_purge_all(void) {
	int i;
	for (i=0; i<num_pools; i++)
		pool_purge(i);
	num_pools = 0;
}
